@@ -4,7 +4,7 @@
  */
 
 /*
- * Copyright (C) 2010-2015 Oracle Corporation
+ * Copyright (C) 2010-2020 Oracle Corporation
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
@@ -17,9 +17,10 @@
 
 #define SCM_WITH_DYNAMIC_LIB_SVN
 
-/*******************************************************************************
-*   Header Files                                                               *
-*******************************************************************************/
+
+/*********************************************************************************************************************************
+*   Header Files                                                                                                                 *
+*********************************************************************************************************************************/
 #include <iprt/assert.h>
 #include <iprt/ctype.h>
 #include <iprt/dir.h>
@@ -47,9 +48,9 @@
 #endif
 
 
-/*******************************************************************************
-*   Defined Constants And Macros                                               *
-*******************************************************************************/
+/*********************************************************************************************************************************
+*   Defined Constants And Macros                                                                                                 *
+*********************************************************************************************************************************/
 #ifdef SCM_WITH_DYNAMIC_LIB_SVN
 # if defined(RT_OS_WINDOWS) && defined(RT_ARCH_X86)
 #  define APR_CALL                       __stdcall
@@ -65,9 +66,9 @@
 #endif
 
 
-/*******************************************************************************
-*   Structures and Typedefs                                                    *
-*******************************************************************************/
+/*********************************************************************************************************************************
+*   Structures and Typedefs                                                                                                      *
+*********************************************************************************************************************************/
 #if defined(SCM_WITH_DYNAMIC_LIB_SVN) && !defined(SCM_WITH_SVN_HEADERS)
 typedef int                         apr_status_t;
 typedef int64_t                     apr_time_t;
@@ -123,9 +124,9 @@ typedef enum svn_depth_t
 #endif /* SCM_WITH_DYNAMIC_LIB_SVN && !SCM_WITH_SVN_HEADERS */
 
 
-/*******************************************************************************
-*   Global Variables                                                           *
-*******************************************************************************/
+/*********************************************************************************************************************************
+*   Global Variables                                                                                                             *
+*********************************************************************************************************************************/
 static char g_szSvnPath[RTPATH_MAX];
 static enum
 {
@@ -158,6 +159,22 @@ static svn_error_t *         (SVN_CALL *g_pfnSvnClientPropGet4)(apr_hash_t **ppH
                                                                 svn_client_ctx_t *pCtx, apr_pool_t *pResultPool,
                                                                 apr_pool_t *pScratchPool);
 /**@} */
+
+/** Cached APR pool. */
+static apr_pool_t          *g_pSvnPool = NULL;
+/** Cached SVN client context. */
+static svn_client_ctx_t    *g_pSvnClientCtx = NULL;
+/** Number of times the current context has been used. */
+static uint32_t             g_cSvnClientCtxUsed = 0;
+
+#endif
+
+
+/*********************************************************************************************************************************
+*   Internal Functions                                                                                                           *
+*********************************************************************************************************************************/
+#ifdef SCM_WITH_DYNAMIC_LIB_SVN
+static void scmSvnFlushClientContextAndPool(void);
 #endif
 
 
@@ -392,15 +409,16 @@ int RTProcExecToString(const char *pszExec, const char * const *papszArgs, RTENV
              * Create the process.
              */
             RTPROCESS hProc;
-            rc = RTProcCreateEx(g_szSvnPath,
+            rc = RTProcCreateEx(pszExec,
                                 papszArgs,
-                                RTENV_DEFAULT,
+                                hEnv,
                                 0 /*fFlags*/,
                                 NULL /*phStdIn*/,
                                 phChildStdOut,
                                 phChildStdErr,
                                 NULL /*pszAsUser*/,
                                 NULL /*pszPassword*/,
+                                NULL /*pvExtraData*/,
                                 &hProc);
             rc2 = RTHandleClose(&hChildStdErr); AssertRC(rc2);
             rc2 = RTHandleClose(&hChildStdOut); AssertRC(rc2);
@@ -527,17 +545,18 @@ int RTProcExec(const char *pszExec, const char * const *papszArgs, RTENV hEnv, u
     /*
      * Create the process.
      */
-    RTPROCESS hProc;
+    RTPROCESS hProc = NIL_RTPROCESS;
     if (RT_SUCCESS(rc))
-        rc = RTProcCreateEx(g_szSvnPath,
+        rc = RTProcCreateEx(pszExec,
                             papszArgs,
-                            RTENV_DEFAULT,
+                            hEnv,
                             0 /*fFlags*/,
                             aph[0],
                             aph[1],
                             aph[2],
                             NULL /*pszAsUser*/,
                             NULL /*pszPassword*/,
+                            NULL /*pvExtraData*/,
                             &hProc);
 
     for (uint32_t i = 0; i < 3; i++)
@@ -565,6 +584,10 @@ static int scmSvnRunAndGetOutput(PSCMRWSTATE pState, const char **papszArgs, boo
 {
     *ppszStdOut = NULL;
 
+#ifdef SCM_WITH_DYNAMIC_LIB_SVN
+    scmSvnFlushClientContextAndPool();
+#endif
+
     char *pszCmdLine = NULL;
     int rc = RTGetOptArgvToString(&pszCmdLine, papszArgs, RTGETOPTARGV_CNV_QUOTE_BOURNE_SH);
     if (RT_FAILURE(rc))
@@ -581,7 +604,7 @@ static int scmSvnRunAndGetOutput(PSCMRWSTATE pState, const char **papszArgs, boo
     {
         if (fNormalFailureOk || Status.enmReason != RTPROCEXITREASON_NORMAL)
             RTMsgError("%s: %s -> %s %u\n",
-                       pszCmdLine,
+                       pState ? pState->pszFilename : "<NONE>", pszCmdLine,
                        Status.enmReason == RTPROCEXITREASON_NORMAL   ? "exit code"
                        : Status.enmReason == RTPROCEXITREASON_SIGNAL ? "signal"
                        : Status.enmReason == RTPROCEXITREASON_ABEND  ? "abnormal end"
@@ -619,13 +642,17 @@ static int scmSvnRunAndGetOutput(PSCMRWSTATE pState, const char **papszArgs, boo
  */
 static int scmSvnRun(PSCMRWSTATE pState, const char **papszArgs, bool fNormalFailureOk)
 {
+#ifdef SCM_WITH_DYNAMIC_LIB_SVN
+    scmSvnFlushClientContextAndPool();
+#endif
+
     char *pszCmdLine = NULL;
     int rc = RTGetOptArgvToString(&pszCmdLine, papszArgs, RTGETOPTARGV_CNV_QUOTE_BOURNE_SH);
     if (RT_FAILURE(rc))
         return rc;
     ScmVerbose(pState, 2, "executing: %s\n", pszCmdLine);
 
-    /* Lazy bird uses RTProcExecToString. */
+    /* Lazy bird uses RTProcExec. */
     RTPROCSTATUS Status;
     rc = RTProcExec(g_szSvnPath, papszArgs, RTENV_DEFAULT, RTPROCEXEC_FLAGS_STD_NULL, &Status);
 
@@ -689,9 +716,23 @@ static void scmSvnTryResolveFunctions(void)
 # else
             { "../lib/lib", ".so" },
             { "../lib/lib", "-1.so" },
+#  if ARCH_BITS == 32
+            { "../lib32/lib", ".so" },
+            { "../lib32/lib", "-1.so" },
+#  else
+            { "../lib64/lib", ".so" },
+            { "../lib64/lib", "-1.so" },
+#  endif
+#  ifdef RT_ARCH_X86
+            { "../lib/i386-linux-gnu/lib", ".so" },
+            { "../lib/i386-linux-gnu/lib", "-1.so" },
+#  else
+            { "../lib/x86_64-linux-gnu/lib", ".so" },
+            { "../lib/x86_64-linux-gnu/lib", "-1.so" },
+#  endif
 # endif
         };
-        for (unsigned iVar = 0; RT_ELEMENTS(s_aVariations); iVar++)
+        for (unsigned iVar = 0; iVar < RT_ELEMENTS(s_aVariations); iVar++)
         {
             /*
              * Try load the svn_client library ...
@@ -703,18 +744,29 @@ static void scmSvnTryResolveFunctions(void)
             unsigned iLib;
             for (iLib = 0; iLib < RT_ELEMENTS(s_apszLibraries) && RT_SUCCESS(rc); iLib++)
             {
-                *pszEndPath = '\0';
-                rc = RTPathAppend(szPath, sizeof(szPath), s_aVariations[iVar].pszPrefix);
-                if (RT_SUCCESS(rc))
-                    rc = RTStrCat(szPath, sizeof(szPath), s_apszLibraries[iLib]);
-                if (RT_SUCCESS(rc))
-                    rc = RTStrCat(szPath, sizeof(szPath), s_aVariations[iVar].pszSuffix);
-                if (RT_SUCCESS(rc))
+                static const char * const s_apszSuffixes[] = { "", ".0", ".1" };
+                for (unsigned iSuff = 0; iSuff < RT_ELEMENTS(s_apszSuffixes); iSuff++)
                 {
+                    *pszEndPath = '\0';
+                    rc = RTPathAppend(szPath, sizeof(szPath), s_aVariations[iVar].pszPrefix);
+                    if (RT_SUCCESS(rc))
+                        rc = RTStrCat(szPath, sizeof(szPath), s_apszLibraries[iLib]);
+                    if (RT_SUCCESS(rc))
+                        rc = RTStrCat(szPath, sizeof(szPath), s_aVariations[iVar].pszSuffix);
+                    if (RT_SUCCESS(rc))
+                        rc = RTStrCat(szPath, sizeof(szPath), s_apszSuffixes[iSuff]);
+                    if (RT_SUCCESS(rc))
+                    {
 # ifdef RT_OS_WINDOWS
-                    RTPathChangeToDosSlashes(pszEndPath, false);
+                        RTPathChangeToDosSlashes(pszEndPath, false);
 # endif
-                    rc = RTLdrLoadEx(szPath, &ahMods[iLib], RTLDRLOAD_FLAGS_NT_SEARCH_DLL_LOAD_DIR , NULL);
+                        rc = RTLdrLoadEx(szPath, &ahMods[iLib], RTLDRLOAD_FLAGS_NT_SEARCH_DLL_LOAD_DIR , NULL);
+                        if (RT_SUCCESS(rc))
+                        {
+                            RTMEM_WILL_LEAK(ahMods[iLib]);
+                            break;
+                        }
+                    }
                 }
             }
             if (iLib == RT_ELEMENTS(s_apszLibraries) && RT_SUCCESS(rc))
@@ -812,11 +864,11 @@ static void scmSvnFindSvnBinary(PSCMRWSTATE pState)
     if (RT_SUCCESS(rc))
     {
         char *pszStripped = RTStrStrip(pszVersion);
-        if (RTStrVersionCompare(pszVersion, "1.8") >= 0)
+        if (RTStrVersionCompare(pszStripped, "1.8") >= 0)
             g_enmSvnVersion = kScmSvnVersion_1_8;
-        else if (RTStrVersionCompare(pszVersion, "1.7") >= 0)
+        else if (RTStrVersionCompare(pszStripped, "1.7") >= 0)
             g_enmSvnVersion = kScmSvnVersion_1_7;
-        else if (RTStrVersionCompare(pszVersion, "1.6") >= 0)
+        else if (RTStrVersionCompare(pszStripped, "1.6") >= 0)
             g_enmSvnVersion = kScmSvnVersion_1_6;
         else
             g_enmSvnVersion = kScmSvnVersion_Ancient;
@@ -907,14 +959,97 @@ static bool scmSvnReadNumber(const char *pch, size_t cch, size_t *pu)
 static int scmSvnAbsPath(const char *pszPath, char *pszAbsPath, size_t cbAbsPath)
 {
     int rc = RTPathAbs(pszPath, pszAbsPath, cbAbsPath);
-# if defined(RT_OS_WINDOWS) || defined(RT_OS_OS2)
+# if RTPATH_STYLE == RTPATH_STR_F_STYLE_DOS
     if (RT_SUCCESS(rc))
+    {
         RTPathChangeToUnixSlashes(pszAbsPath, true /*fForce*/);
+        /* To avoid: svn: E235000: In file '..\..\..\subversion\libsvn_client\prop_commands.c' line 796: assertion failed (svn_dirent_is_absolute(target)) */
+        if (pszAbsPath[1] == ':')
+            pszAbsPath[0] = RT_C_TO_UPPER(pszAbsPath[0]);
+    }
 # endif
     return rc;
 }
 
 
+/**
+ * Gets a client context and pool.
+ *
+ * This implements caching.
+ *
+ * @returns IPRT status code.
+ * @param   ppCtx               Where to return the context
+ * @param   ppPool              Where to return the pool.
+ */
+static int scmSvnGetClientContextAndPool(svn_client_ctx_t **ppCtx, apr_pool_t **ppPool)
+{
+    /*
+     * Use cached if present.
+     */
+    if (g_pSvnClientCtx && g_pSvnPool)
+    {
+        g_cSvnClientCtxUsed++;
+        *ppCtx  = g_pSvnClientCtx;
+        *ppPool = g_pSvnPool;
+        return VINF_SUCCESS;
+    }
+    Assert(!g_pSvnClientCtx);
+    Assert(!g_pSvnPool);
+
+    /*
+     * Create new pool and context.
+     */
+    apr_pool_t *pPool = g_pfnSvnPoolCreateEx(NULL, NULL);
+    if (pPool)
+    {
+        svn_client_ctx_t *pCtx = NULL;
+        svn_error_t *pErr = g_pfnSvnClientCreateContext(&pCtx, pPool);
+        if (!pErr)
+        {
+            g_cSvnClientCtxUsed = 1;
+            g_pSvnClientCtx     = *ppCtx  = pCtx;
+            g_pSvnPool          = *ppPool = pPool;
+            return VINF_SUCCESS;
+        }
+        g_pfnAprPoolDestroy(pPool);
+    }
+
+    *ppCtx  = NULL;
+    *ppPool = NULL;
+    return VERR_GENERAL_FAILURE;
+}
+
+
+/**
+ * Puts back a client context and pool after use.
+ *
+ * @param   pCtx                The context.
+ * @param   pPool               The pool.
+ * @param   fFlush              Whether to flush it.
+ */
+static void scmSvnPutClientContextAndPool(svn_client_ctx_t *pCtx, apr_pool_t *pPool, bool fFlush)
+{
+    if (fFlush || g_cSvnClientCtxUsed > 4096) /* Disable this to force new context every time. */
+    {
+        g_pfnAprPoolDestroy(pPool);
+        g_pSvnPool = NULL;
+        g_pSvnClientCtx = NULL;
+    }
+    RT_NOREF(pCtx, fFlush);
+}
+
+
+/**
+ * Flushes the cached client context and pool
+ */
+static void scmSvnFlushClientContextAndPool(void)
+{
+    if (g_pSvnPool)
+        scmSvnPutClientContextAndPool(g_pSvnClientCtx, g_pSvnPool, true /*fFlush*/);
+    Assert(!g_pSvnPool);
+}
+
+
 /**
  * Checks if @a pszPath exists in the current WC.
  *
@@ -923,35 +1058,31 @@ static int scmSvnAbsPath(const char *pszPath, char *pszAbsPath, size_t cbAbsPath
  */
 static int scmSvnIsObjectInWorkingCopy(const char *pszPath)
 {
-    int rc = -1;
-
     /* svn_client_propget4 and later requires absolute target path. */
     char szAbsPath[RTPATH_MAX];
-    int  rc2 = scmSvnAbsPath(pszPath, szAbsPath, sizeof(szAbsPath));
-    if (RT_SUCCESS(rc2))
+    int  rc = scmSvnAbsPath(pszPath, szAbsPath, sizeof(szAbsPath));
+    if (RT_SUCCESS(rc))
     {
-        /* Create calling context. */
-        apr_pool_t *pPool = g_pfnSvnPoolCreateEx(NULL, NULL);
-        if (pPool)
+        apr_pool_t *pPool;
+        svn_client_ctx_t *pCtx = NULL;
+        rc = scmSvnGetClientContextAndPool(&pCtx, &pPool);
+        if (RT_SUCCESS(rc))
         {
-            svn_client_ctx_t *pCtx = NULL;
-            svn_error_t *pErr = g_pfnSvnClientCreateContext(&pCtx, pPool);
+            /* Make the call. */
+            apr_hash_t         *pHash = NULL;
+            svn_opt_revision_t  Rev;
+            RT_ZERO(Rev);
+            Rev.kind          = svn_opt_revision_working;
+            Rev.value.number  = -1L;
+            svn_error_t *pErr = g_pfnSvnClientPropGet4(&pHash, "svn:no-such-property", szAbsPath, &Rev, &Rev,
+                                                       NULL /*pActualRev*/, svn_depth_empty, NULL /*pChangeList*/,
+                                                       pCtx, pPool, pPool);
             if (!pErr)
-            {
-                /* Make the call. */
-                apr_hash_t         *pHash = NULL;
-                svn_opt_revision_t  Rev;
-                RT_ZERO(Rev);
-                Rev.kind          = svn_opt_revision_base;
-                Rev.value.number  = -1L;
-                pErr = g_pfnSvnClientPropGet4(&pHash, "svn:no-such-property", szAbsPath, &Rev, &Rev,
-                                              NULL /*pActualRev*/, svn_depth_empty, NULL /*pChangeList*/, pCtx, pPool, pPool);
-                if (!pErr)
-                    rc = true;
-                else if (pErr->apr_err == SVN_ERR_UNVERSIONED_RESOURCE)
-                    rc = false;
-            }
-            g_pfnAprPoolDestroy(pPool);
+                rc = true;
+            else if (pErr->apr_err == SVN_ERR_UNVERSIONED_RESOURCE)
+                rc = false;
+
+            scmSvnPutClientContextAndPool(pCtx, pPool, false);
         }
     }
     return rc;
@@ -964,18 +1095,26 @@ static int scmSvnIsObjectInWorkingCopy(const char *pszPath)
  * Checks if the file we're operating on is part of a SVN working copy.
  *
  * @returns true if it is, false if it isn't or we cannot tell.
- * @param   pState              The rewrite state to work on.
+ * @param   pState      The rewrite state to work on.  Will use the
+ *                      fIsInSvnWorkingCopy member for caching the result.
  */
 bool ScmSvnIsInWorkingCopy(PSCMRWSTATE pState)
 {
-    scmSvnFindSvnBinary(pState);
+    /*
+     * We don't ask SVN twice as that's expensive.
+     */
+    if (pState->fIsInSvnWorkingCopy != 0)
+        return pState->fIsInSvnWorkingCopy > 0;
 
 #ifdef SCM_WITH_DYNAMIC_LIB_SVN
     if (g_fSvnFunctionPointersValid)
     {
         int rc = scmSvnIsObjectInWorkingCopy(pState->pszFilename);
         if (rc == (int)true || rc == (int)false)
+        {
+            pState->fIsInSvnWorkingCopy = rc == (int)true ? 1 : -1;
             return rc == (int)true;
+        }
     }
 
     /* Fallback: */
@@ -988,19 +1127,27 @@ bool ScmSvnIsInWorkingCopy(PSCMRWSTATE pState)
         char szPath[RTPATH_MAX];
         int rc = scmSvnConstructName(pState, ".svn/text-base/", ".svn-base", szPath);
         if (RT_SUCCESS(rc))
-            return RTFileExists(szPath);
+        {
+            if (RTFileExists(szPath))
+            {
+                pState->fIsInSvnWorkingCopy = 1;
+                return true;
+            }
+        }
     }
     else
     {
-        const char *apszArgs[] = { g_szSvnPath, "propget", "svn:no-such-property", pState->pszFilename, NULL };
+        const char *apszArgs[] = { g_szSvnPath, "proplist", pState->pszFilename, NULL };
         char       *pszValue;
         int rc = scmSvnRunAndGetOutput(pState, apszArgs, true, &pszValue);
         if (RT_SUCCESS(rc))
         {
             RTStrFree(pszValue);
+            pState->fIsInSvnWorkingCopy = 1;
             return true;
         }
     }
+    pState->fIsInSvnWorkingCopy = -1;
     return false;
 }
 
@@ -1013,8 +1160,6 @@ bool ScmSvnIsInWorkingCopy(PSCMRWSTATE pState)
  */
 bool ScmSvnIsDirInWorkingCopy(const char *pszDir)
 {
-    scmSvnFindSvnBinary(NULL);
-
 #ifdef SCM_WITH_DYNAMIC_LIB_SVN
     if (g_fSvnFunctionPointersValid)
     {
@@ -1061,52 +1206,48 @@ bool ScmSvnIsDirInWorkingCopy(const char *pszDir)
  */
 static int scmSvnQueryPropertyUsingApi(const char *pszPath, const char *pszProperty, char **ppszValue)
 {
-    int rc = VERR_NOT_SUPPORTED;
-
     /* svn_client_propget4 and later requires absolute target path. */
     char szAbsPath[RTPATH_MAX];
-    int  rc2 = scmSvnAbsPath(pszPath, szAbsPath, sizeof(szAbsPath));
-    if (RT_SUCCESS(rc2))
+    int  rc = scmSvnAbsPath(pszPath, szAbsPath, sizeof(szAbsPath));
+    if (RT_SUCCESS(rc))
     {
-        /* Create calling context. */
-        apr_pool_t *pPool = g_pfnSvnPoolCreateEx(NULL, NULL);
-        if (pPool)
+        apr_pool_t *pPool;
+        svn_client_ctx_t *pCtx = NULL;
+        rc = scmSvnGetClientContextAndPool(&pCtx, &pPool);
+        if (RT_SUCCESS(rc))
         {
-            svn_client_ctx_t *pCtx = NULL;
-            svn_error_t *pErr = g_pfnSvnClientCreateContext(&pCtx, pPool);
+            /* Make the call. */
+            apr_hash_t         *pHash = NULL;
+            svn_opt_revision_t  Rev;
+            RT_ZERO(Rev);
+            Rev.kind          = svn_opt_revision_working;
+            Rev.value.number  = -1L;
+            svn_error_t *pErr = g_pfnSvnClientPropGet4(&pHash, pszProperty, szAbsPath, &Rev, &Rev,
+                                                       NULL /*pActualRev*/, svn_depth_empty, NULL /*pChangeList*/,
+                                                       pCtx, pPool, pPool);
             if (!pErr)
             {
-                /* Make the call. */
-                apr_hash_t         *pHash = NULL;
-                svn_opt_revision_t  Rev;
-                RT_ZERO(Rev);
-                Rev.kind          = svn_opt_revision_base;
-                Rev.value.number  = -1L;
-                pErr = g_pfnSvnClientPropGet4(&pHash, pszProperty, szAbsPath, &Rev, &Rev,
-                                              NULL /*pActualRev*/, svn_depth_empty, NULL /*pChangeList*/, pCtx, pPool, pPool);
-                if (!pErr)
+                /* Get the first value, if any. */
+                rc = VERR_NOT_FOUND;
+                apr_hash_index_t *pHashIdx = g_pfnAprHashFirst(pPool, pHash);
+                if (pHashIdx)
                 {
-                    /* Get the first value, if any. */
-                    rc = VERR_NOT_FOUND;
-                    apr_hash_index_t *pHashIdx = g_pfnAprHashFirst(pPool, pHash);
-                    if (pHashIdx)
+                    const char **ppszFirst = (const char **)g_pfnAprHashThisVal(pHashIdx);
+                    if (ppszFirst && *ppszFirst)
                     {
-                        const char **ppszFirst = (const char **)g_pfnAprHashThisVal(pHashIdx);
-                        if (ppszFirst && *ppszFirst)
-                        {
-                            if (ppszValue)
-                                rc = RTStrDupEx(ppszValue, *ppszFirst);
-                            else
-                                rc = VINF_SUCCESS;
-                        }
+                        if (ppszValue)
+                            rc = RTStrDupEx(ppszValue, *ppszFirst);
+                        else
+                            rc = VINF_SUCCESS;
                     }
                 }
-                else if (pErr->apr_err == SVN_ERR_UNVERSIONED_RESOURCE)
-                    rc = VERR_INVALID_STATE;
-                else
-                    rc = VERR_GENERAL_FAILURE;
             }
-            g_pfnAprPoolDestroy(pPool);
+            else if (pErr->apr_err == SVN_ERR_UNVERSIONED_RESOURCE)
+                rc = VERR_INVALID_STATE;
+            else
+                rc = VERR_GENERAL_FAILURE;
+
+            scmSvnPutClientContextAndPool(pCtx, pPool, false);
         }
     }
     return rc;
@@ -1146,8 +1287,6 @@ int ScmSvnQueryProperty(PSCMRWSTATE pState, const char *pszName, char **ppszValu
             return VINF_SUCCESS;
         }
 
-    scmSvnFindSvnBinary(pState);
-
 #ifdef SCM_WITH_DYNAMIC_LIB_SVN
     if (g_fSvnFunctionPointersValid)
     {
@@ -1307,6 +1446,44 @@ int ScmSvnQueryProperty(PSCMRWSTATE pState, const char *pszName, char **ppszValu
 }
 
 
+/**
+ * Queries the value of an SVN property on the parent dir/whatever.
+ *
+ * This will not adjust for scheduled changes to the parent!
+ *
+ * @returns IPRT status code.
+ * @retval  VERR_INVALID_STATE if not a SVN WC file.
+ * @retval  VERR_NOT_FOUND if the property wasn't found.
+ * @param   pState              The rewrite state to work on.
+ * @param   pszName             The property name.
+ * @param   ppszValue           Where to return the property value.  Free this
+ *                              using RTStrFree.  Optional.
+ */
+int ScmSvnQueryParentProperty(PSCMRWSTATE pState, const char *pszName, char **ppszValue)
+{
+    /*
+     * Strip the filename and use ScmSvnQueryProperty.
+     */
+    char szPath[RTPATH_MAX];
+    int rc = RTStrCopy(szPath, sizeof(szPath), pState->pszFilename);
+    if (RT_SUCCESS(rc))
+    {
+        RTPathStripFilename(szPath);
+        SCMRWSTATE ParentState;
+        ParentState.pszFilename         = szPath;
+        ParentState.fFirst              = false;
+        ParentState.fIsInSvnWorkingCopy = true;
+        ParentState.cSvnPropChanges     = 0;
+        ParentState.paSvnPropChanges    = NULL;
+        ParentState.rc                  = VINF_SUCCESS;
+        rc = ScmSvnQueryProperty(&ParentState, pszName, ppszValue);
+        if (RT_SUCCESS(rc))
+            rc = ParentState.rc;
+    }
+    return rc;
+}
+
+
 /**
  * Schedules the setting of a property.
  *
@@ -1415,8 +1592,6 @@ int ScmSvnDisplayChanges(PSCMRWSTATE pState)
  */
 int ScmSvnApplyChanges(PSCMRWSTATE pState)
 {
-    scmSvnFindSvnBinary(pState);
-
 #ifdef SCM_WITH_LATER
     if (0)
     {
@@ -1449,3 +1624,19 @@ int ScmSvnApplyChanges(PSCMRWSTATE pState)
     return VINF_SUCCESS;
 }
 
+
+/**
+ * Initializes the subversion interface.
+ */
+void ScmSvnInit(void)
+{
+    scmSvnFindSvnBinary(NULL);
+}
+
+
+void ScmSvnTerm(void)
+{
+#ifdef SCM_WITH_DYNAMIC_LIB_SVN
+    scmSvnFlushClientContextAndPool();
+#endif
+}
